(ns propeller.gp
  (:require [clojure.string]
            [clojure.pprint]
            [propeller.genome :as genome]
            [propeller.variation :as variation]
            [propeller.push.instructions.bool]
            [propeller.push.instructions.character]
            [propeller.push.instructions.code]
            [propeller.push.instructions.input-output]
            [propeller.push.instructions.numeric]
            [propeller.push.instructions.polymorphic]
            [propeller.push.instructions.string]
            [propeller.push.instructions.vector]
            [psb2.core :as psb2]))

(defn report
  "Reports information each generation."
  [pop generation argmap]
  (let [best (first pop)]
    (println {:generation            generation
              :best-plushy           (:plushy best)
              :best-program          (genome/plushy->push (:plushy best) argmap)
              :best-total-error      (:total-error best)
              :best-errors           (:errors best)
              :best-behaviors        (:behaviors best)
              :genotypic-diversity   (float (/ (count (distinct (map :plushy pop))) (count pop)))
              :behavioral-diversity  (float (/ (count (distinct (map :behaviors pop))) (count pop)))
              :average-genome-length (float (/ (reduce + (map count (map :plushy pop))) (count pop)))
              :average-total-error   (float (/ (reduce + (map :total-error pop)) (count pop)))})
    (println)))
;  (clojure.pprint/pprint


(defn gp
  "Main GP loop."
  [{:keys [population-size max-generations error-function instructions
           max-initial-plushy-size PSB2-path PSB2-problem]
    :as   argmap}]
  ;;
  (println {:starting-args argmap})
  (println)
  ;;
  (let [PSB2-data (if (= PSB2-path "")
                    #{}
                    (psb2/fetch-examples PSB2-path PSB2-problem 200 2000))
        enhanced-argmap (assoc argmap :train-and-test-data PSB2-data)]

    (loop [generation 0
           population (repeatedly
                        population-size
                        #(hash-map :plushy (genome/make-random-plushy
                                             instructions
                                             max-initial-plushy-size)))]
      (let [evaluated-pop (sort-by :total-error
                                   (#?(:clj  pmap
                                       :cljs map)
                                     (partial error-function enhanced-argmap) population))
            best-individual (first evaluated-pop)]
        (report evaluated-pop generation argmap)
        (cond
          ;; Success on training cases is verified on testing cases
          (zero? (:total-error best-individual))
          (do (println {:success-generation generation})
              (println {:total-test-error (:total-error (error-function argmap best-individual :test))})
              (#?(:clj shutdown-agents)))
          ;;
          (>= generation max-generations)
          nil
          ;;
          :else (recur (inc generation)
                       (if (:elitism argmap)
                         (conj (repeatedly (dec population-size)
                                           #(variation/new-individual evaluated-pop argmap))
                               (first evaluated-pop))
                         (repeatedly population-size
                                     #(variation/new-individual evaluated-pop argmap)))))))))